#include "../../config.h"

#ifndef __SCROBBLER_CONFIG_H__
#define __SCROBBLER_CONFIG_H__

#define DEBUG 0
#define META_DEBUG 0
#define SUB_DEBUG 0
#define CLIENT "Audacious"
#define USER_AGENT "AudioScrobbler/1.1" PACKAGE_NAME "/" PACKAGE_VERSION
#define MAKE_BMP
#define ALLOW_MULTIPLE

#endif
