/*
 * Utils for libavcodec
 * Copyright (c) 2002 Fabrice Bellard.
 * Copyright (c) 2004 Roman Bogorodskiy (bmp-wma specific stuff)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "avcodec.h"

void avcodec_register_all(void)
{
	static int inited = 0;
    
    	if (inited != 0)
		return;
    	inited = 1;

    	register_avcodec(&wmav1_decoder);
    	register_avcodec(&wmav2_decoder);
}

