/*
*
* Author: Giacomo Lozito <james@develia.org>, (C) 2005-2006
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2 of the License, or (at your
* option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*
*/

#ifndef _I_CONFIGURE_H
#define _I_CONFIGURE_H 1

#include "i_common.h"


typedef struct
{
  gchar *	ap_seq_backend;
  gint		ap_opts_transpose_value;
  gint		ap_opts_drumshift_value;
  gint		ap_opts_length_precalc;
  gint		ap_opts_comments_extract;
  gint		ap_opts_lyrics_extract;
}
amidiplug_cfg_ap_t;


extern amidiplug_cfg_ap_t amidiplug_cfg_ap;


void i_configure_gui( void );
void i_configure_cfg_ap_read( void );
void i_configure_cfg_ap_save( void );


#endif /* !_I_CONFIGURE_H */
